/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Picture
Version:   1.0
Include:   pandora/graphics/picture.h
Status:    Stable
ID:        ID_PICTURE
Category:  Graphics
Date:      October 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 1996-2003. All rights reserved.
Keywords:  picture, image, bitmap, jpeg, gif, iff, pcx, png
Short:     Loads and saves picture files in a variety of different data formats.

###DESCRIPTION###
<p>The purpose of the Picture class is to provide a standard interface that
any program can use to load picture files.  The main advantage in using this
class is that the source picture can be in any file format recognised by the
user's system, so effectively your program will be able to support file
formats that it does not understand.</p>

<p>The default file format for loading and saving pictures is IFF, the other
formats (such as JPEG) are supported via sub-classes, which can be loaded into
the system at boot time or on demand.  All Picture sub-classes are available
from the Rocklyte FTP file repository, which you can check if you need to know
what file formats are currently supported.</p>

###TECHNICAL###
<p>The Picture class will clip any loaded picture so that it fits the size
given in the <a href="#Bitmap">Bitmap</a>'s Width and Height. If you
specify the RESIZE <a href="#Flags">flag</a>, the
picture will be shrunk or enlarged to fit the given dimensions. If you
leave the Width and Height at NULL, then the picture will be loaded at its
default dimensions.  If you need to find out general information about a
picture before initialising it, you may <a href="#Query">Query</a> it first so that the picture
object can load initial details on the file format.</p>

<p>Images are also remapped automatically if the source palette and
destination palettes do not match, or if there are significant differences
between the source and destination bitmap types.</p>

###END###

*****************************************************************************/

#define LOCAL_PICTURES
#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/file.h>
#include <pandora/modules/filesystem.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_PICTURE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1996-2003.  All rights reserved.",
   "October 2003",
   "Picture"
};

static struct KernelBase *KernelBase = NULL;
static OBJECTPTR PictureClass = NULL;
static struct ModuleMaster *PictureModule = NULL;
static OBJECTPTR FileSystemMod = NULL;
static struct FileSystemBase *FileSystemBase;

static ERROR MODULE_ActionNotify(OBJECTPTR, struct acActionNotify *Args);

#define UNPACKSIZE 8192
#define BUFFERSIZE 4096

#define OSV_HIRES (1L<<15)
#define OSV_HAM   (1L<<11)
#define OSV_EHB   (1L<<10)
#define OSV_LACED (1L<<2)

struct BMHD {
   WORD  Width;       /* Picture width */
   WORD  Height;      /* Picture height */
   WORD  X;           /* ? */
   WORD  Y;           /* ? */
   BYTE  Depth;       /* Amount of planes */
   BYTE  Mask;        /* Masking technique in use */
   BYTE  Pack;        /* 1 if ByteRunOne, 0 if none */
   BYTE  PAD;         /* Empty */
   WORD  TColor;      /* Transparent colour number */
   BYTE  XAspect;     /* Pixel Width */
   BYTE  YAspect;     /* Height ratio */
   WORD  ScrWidth;    /* Screen/ViewPort Width */
   WORD  ScrHeight;   /* Screen/ViewPort Height */
};

static LONG UnpackPicture(struct Picture *, struct BMHD *, OBJECTPTR,  LONG, BYTE *, LONG);
static BYTE * SkipLine(struct BMHD *, BYTE *, OBJECTPTR, struct Bitmap *);
static BYTE * UnpackPlane(struct BMHD *, OBJECTPTR, struct Bitmap *, BYTE *, BYTE *);

static ERROR PIC_Activate(struct Picture *, APTR);
static ERROR PIC_AccessObject(struct Picture *, APTR);
static ERROR PIC_Free(struct Picture *, APTR);
static ERROR PIC_ReleaseObject(struct Picture *, APTR);
static ERROR PIC_NewObject(struct Picture *, APTR);
static ERROR PIC_Init(struct Picture *, APTR);
static ERROR PIC_Query(struct Picture *, APTR);
static ERROR PIC_Read(struct Picture *, struct acRead *);
static ERROR PIC_SaveImage(struct Picture *, struct acSaveImage *);
static ERROR PIC_SaveToObject(struct Picture *, struct acSaveToObject *);
static ERROR PIC_Seek(struct Picture *, struct acSeek *);
static ERROR PIC_Write(struct Picture *, struct acWrite *);

static ERROR GET_Author(struct Picture *, STRING *);
static ERROR GET_Copyright(struct Picture *, STRING *);
static ERROR GET_Description(struct Picture *, STRING *);
static ERROR GET_Disclaimer(struct Picture *, STRING *);
static ERROR GET_Header(struct Picture *, APTR *);
static ERROR GET_Location(struct Picture *, STRING *);
static ERROR GET_Software(struct Picture *, STRING *);
static ERROR GET_Title(struct Picture *, STRING *);

static ERROR SET_Author(struct Picture *, STRING);
static ERROR SET_Copyright(struct Picture *, STRING);
static ERROR SET_Description(struct Picture *, STRING);
static ERROR SET_Disclaimer(struct Picture *, STRING);
static ERROR SET_Location(struct Picture *, STRING);
static ERROR SET_Software(struct Picture *, STRING);
static ERROR SET_Title(struct Picture *, STRING);

static struct FieldDef PicFlags[] = {
   { "RESIZEX",   PCF_RESIZEX   }, { "RESIZEY", PCF_RESIZEY },
   { "RESIZE",    PCF_RESIZE    }, { "STRETCH", PCF_RESIZE }, { "NEW",   PCF_NEW },
   { "INTERLACE", PCF_INTERLACE }, { "ALPHA",   PCF_ALPHA  }, { "CACHE", PCF_CACHE },
   { NULL, NULL }
};

static struct FieldArray PictureFields[] = {
   { "Bitmap",        0, FDF_CHILD|FDF_R,      ID_BITMAP,       NULL, NULL },
   { "Flags",         0, FDF_LONGFLAGS|FDF_RW, (LONG)&PicFlags, NULL, NULL },
   { "DisplayHeight", 0, FDF_LONG|FDF_RW,      0,               NULL, NULL },
   { "DisplayWidth",  0, FDF_LONG|FDF_RW,      0,               NULL, NULL },
   { "Quality",       0, FDF_LONG|FDF_RW,      0,               NULL, NULL },
   { "Mask",          0, FDF_CHILD|FDF_R,      ID_BITMAP,       NULL, NULL },
   /*** Virtual fields ***/
   { "Author",        0, FDF_STRING|FDF_RW,  0, GET_Author,      SET_Author },
   { "Copyright",     0, FDF_STRING|FDF_RW,  0, GET_Copyright,   SET_Copyright },
   { "Description",   0, FDF_STRING|FDF_RW,  0, GET_Description, SET_Description },
   { "Disclaimer",    0, FDF_STRING|FDF_RW,  0, GET_Disclaimer,  SET_Disclaimer },
   { "Header",        0, FDF_POINTER|FDF_RI, 0, GET_Header,      NULL },
   { "Location",      0, FDF_STRING|FDF_RI,  0, GET_Location,    SET_Location },
   { "Software",      0, FDF_STRING|FDF_RW,  0, GET_Software,    SET_Software },
   { "Title",         0, FDF_STRING|FDF_RW,  0, GET_Title,       SET_Title },
   END_FIELD
};

static struct ActionArray PictureActions[] = {
   { AC_AccessObject,  PIC_AccessObject },
   { AC_Activate,      PIC_Activate },
   { AC_Free,          PIC_Free },
   { AC_NewObject,     PIC_NewObject },
   { AC_Init,          PIC_Init },
   { AC_Query,         PIC_Query },
   { AC_Read,          PIC_Read },
   { AC_ReleaseObject, PIC_ReleaseObject },
   { AC_SaveImage,     PIC_SaveImage },
   { AC_SaveToObject,  PIC_SaveToObject },
   { AC_Seek,          PIC_Seek },
   { AC_Write,         PIC_Write },
   { NULL, NULL }
};

/****************************************************************************/

static ERROR PIC_CacheBitmap(struct Picture *, APTR);

static struct MethodArray PictureMethods[] = {
   { MT_CacheBitmap, PIC_CacheBitmap, "CacheBitmap", NULL, NULL },
   { NULL, NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Cache information.
*/

#define CACHE_SIZELIMIT 512 //1024768  /* The picture cache has a limit of this many bytes */

struct PictureCache {
   OBJECTID BitmapID;
   OBJECTID MaskID;
   LARGE Time;            /* Indicates the SystemTime() at which the cache entry was last used */
   LARGE FileTime;        /* Timestamp on the original file */
   LARGE FileSize;        /* Size of the original file */
   LONG  Size;            /* The total number of bytes used to create the raw bitmap data */
   UBYTE Location[200];
   WORD  OpenCount;
   WORD  Year;
   WORD  Month;
   WORD  Day;
   WORD  Hour;
   WORD  Minute;
   WORD  Second;
};

static struct PictureCache *glCache = 0;
static LONG glCacheSize = 0;
static LONG glMemUsed = 0; /* The total amount of bitmap image memory used by the cache system */

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   APTR *actions;

   KernelBase = argKernelBase;

   /* Listen to the task's Free action, for the purpose
   ** of killing the bitmap cache on shutdown.
   */

   if (GetField(argModule, FID_Master, FT_POINTER, &PictureModule) IS ERR_Okay) {
      SubscribeAction(CurrentTask(), AC_Free, PictureModule);
      if (GetField(PictureModule, FID_Actions, FT_POINTER, &actions) IS ERR_Okay) {
         actions[AC_ActionNotify] = MODULE_ActionNotify;
      }
   }
   else return(ERR_GetField);

   if (CreateObject(ID_MODULE, NULL, &FileSystemMod, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(FileSystemMod, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&PictureClass, NULL,
      FID_BaseClassID|TLONG,       ID_PICTURE,
      FID_Version|TFLOAT,          VER_PICTURE,
      FID_FileExtension|TSTRING,   "iff;ilbm;pic",
      FID_FileDescription|TSTRING, "IFF Picture",
      FID_Name|TSTRING,            "Picture",
      FID_Category|TLONG,          CCF_GRAPHICS,
      FID_Actions|TPTR,            PictureActions,
      FID_Methods|TPTR,            PictureMethods,
      FID_Fields|TPTR,             PictureFields,
      FID_Size|TLONG,              sizeof(struct Picture),
      TAGEND));
}

/*****************************************************************************
** Module: ActionNotify
*/

static ERROR MODULE_ActionNotify(OBJECTPTR Module, struct acActionNotify *Args)
{
   WORD i;

   if (Args->ActionID IS AC_Free) {
      /* When the task is shutting down, this routine will destroy all
      ** cached bitmaps.  This ensures that we maintain a clean shutdown
      ** process and we won't get debug messages about unfreed objects.
      */

      DPrintF("Module:","[Picture] Destroying cached bitmaps...");

      for (i=0; i < glCacheSize; i++) {
         if (glCache[i].BitmapID) {
            ActionMsg(AC_Free, glCache[i].BitmapID, NULL);
            if (glCache[i].MaskID) ActionMsg(AC_Free, glCache[i].MaskID, NULL);

            glMemUsed -= glCache[i].Size;
            ClearMemory(glCache + i, sizeof(glCache[i]));
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (PictureClass)  { Action(AC_Free, PictureClass, NULL);  PictureClass = NULL; }
   if (FileSystemMod) { Action(AC_Free, FileSystemMod, NULL); FileSystemMod = NULL; }

   /* NB: Individual cached bitmaps are not removed during expunge, because
   ** the task's shutdown procedure will automatically destroy cached
   ** pictures well before it calls our CMDExpunge() routine.
   */

   if (glCache) FreeMemory(glCache);

   return(ERR_Okay);
}

/*****************************************************************************
** Picture: AccessObject
*/

static ERROR PIC_AccessObject(struct Picture *Self, APTR Void)
{
   if (Self->prvBitmapID) {
      if (AccessObject(Self->prvBitmapID, 0, (OBJECTPTR *)&Self->Bitmap) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }

   if (Self->prvMaskID) {
      if (AccessObject(Self->prvMaskID, 0, (OBJECTPTR *)&Self->Mask) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Loads image data into a picture object.

###DESCRIPTION###
<p>In order to load an image file into a picture object you will need to
Activate it after initialisation.  So long as the <a href="#Location">Location</a>
field refers to a recognised image file, it will be loaded into the picture
object and the fields will be filled out to reflect the image content.</p>

<p>If you have preset the values of certain fields prior to activation, you
will be placing restrictions on the image file that is to be loaded.  For
example, if the source image is wider than a restricted Bitmap Width, the
image will have its right edge clipped.  The same is true for the Bitmap
Height and other restrictions apply to fields such as the Bitmap Palette.</p>

<p>Once the picture is loaded, the image data will be held in the picture's
Bitmap object.  You can draw to and from the Bitmap using its available
drawing methods.</p>

###END###

*****************************************************************************/

static ERROR PIC_Activate(struct Picture *Self, APTR Void)
{
   struct BMHD *BMHD;
   struct acSeek seek;
   struct acRead read;
   LONG BodyPos, CAMG, *CMAP, error, FileSize, ChunkSize, bodylen;
   UBYTE *Buffer;
   BYTE  *BODY;

   /*** Return if the picture object has already been activated ***/

   if (Self->Bitmap->Head.ObjectFlags & NF_INITIALISED) return(ERR_Okay);

   DPrintF("~Activate()","[IFF:%d]", Self->Head.UniqueID);

   BMHD    = 0;
   BodyPos = 0;
   CAMG    = 0;
   CMAP    = 0;
   BODY    = 0;
   Buffer  = 0;
   bodylen = 0;
   error   = ERR_Failed;

   if (AllocMemory(BUFFERSIZE, MEM_DATA|MEM_NOCLEAR, (APTR *)&Buffer, NULL) != ERR_Okay) {
      error = ObjectError(Self, ERH_Init, ERR_AllocMemory);
      goto exit;
   }

   /*** Check for the initial FORM/ILBM part ***/

   ActionTags(AC_Seek, Self->prvFile, 0.0, POS_START);
   SetRead(read, Buffer, 12);
   Action(AC_Read, Self->prvFile, &read);   /* [FORM] [Size] [ILBM] */
   if ((StrCompare(Buffer,"FORM", 4, STR_CASE) != ERR_Okay) OR (StrCompare(Buffer+8,"ILBM", 4, STR_CASE) != ERR_Okay)) {
      DPrintF("Picture:","Not an IFF (FORM/ILBM) picture file.");
      error = ERR_NoSupport;
      goto exit;
   }

   FileSize = rd_long(Buffer + 4) - 4;

   /*** Read each iff chunk in the file ***/

   while ((FileSize > 0) AND (BodyPos IS NULL)) {

      SetRead(read, Buffer, 8);
      Action(AC_Read, Self->prvFile, &read);
      if (read.Result != 8) { /* [CODE] [Size] */
         DPrintF("@Picture:","I could not read the next chunk.");
         goto exit;
      }

      FileSize -= 8;

      ChunkSize = rd_long(Buffer+4);
      if (ChunkSize & 0x00000001) ChunkSize++;

      FileSize -= ChunkSize;

      if (StrCompare(Buffer, "BMHD", 4, STR_CASE) IS ERR_Okay) {
         DPrintF("4Picture:","IFF BMHD header found [%ld].", ChunkSize);
         if (ChunkSize > BUFFERSIZE) goto exit;
         if (AllocMemory(ChunkSize, MEM_NOCLEAR, (APTR *)&BMHD, NULL) IS ERR_Okay) {
            struct mtReadFileWord Args;
            Action(MT_ReadFileWord, Self->prvFile, &Args);     BMHD->Width  = Args.Result;
            Action(MT_ReadFileWord, Self->prvFile, &Args);     BMHD->Height = Args.Result;
            Action(MT_ReadFileWord, Self->prvFile, &Args);     BMHD->X = Args.Result;
            Action(MT_ReadFileWord, Self->prvFile, &Args);     BMHD->Y = Args.Result;
            SetRead(read, &BMHD->Depth, 1);   Action(AC_Read, Self->prvFile, &read);
            SetRead(read, &BMHD->Mask, 1);    Action(AC_Read, Self->prvFile, &read);
            SetRead(read, &BMHD->Pack, 1);    Action(AC_Read, Self->prvFile, &read);
            SetRead(read, &BMHD->PAD, 1);     Action(AC_Read, Self->prvFile, &read);
            Action(MT_ReadFileWord, Self->prvFile, &Args);     BMHD->TColor = Args.Result;
            SetRead(read, &BMHD->XAspect, 1); Action(AC_Read, Self->prvFile, &read);
            SetRead(read, &BMHD->YAspect, 1); Action(AC_Read, Self->prvFile, &read);
            Action(MT_ReadFileWord, Self->prvFile, &Args);     BMHD->ScrWidth  = Args.Result;
            Action(MT_ReadFileWord, Self->prvFile, &Args);     BMHD->ScrHeight = Args.Result;
         }
         else {
            error = ObjectError(Self, ERH_Init, ERR_Memory);
            goto exit;
         }
      }
      else if (StrCompare(Buffer, "CAMG", 4, STR_CASE) IS ERR_Okay) {
         DPrintF("4Picture:","IFF CAMG header found [%ld].", ChunkSize);
         if (ChunkSize > BUFFERSIZE) goto exit;
         SetRead(read, Buffer, ChunkSize);
         Action(AC_Read, Self->prvFile, &read);
         if (read.Result IS ChunkSize) {
            CAMG = rd_long(Buffer);
         }
         else {
            error = ObjectError(Self, ERH_Init, ERR_Read);
            goto exit;
         }
      }
      else if (StrCompare(Buffer, "BODY", 4, STR_CASE) IS ERR_Okay) {
         DPrintF("4Picture:","IFF BODY header found [%ld].", ChunkSize);
         bodylen = ChunkSize;
         if (AllocMemory(bodylen, MEM_NOCLEAR, (APTR *)&BODY, NULL) IS ERR_Okay) {
            SetRead(read, BODY, bodylen);
            Action(AC_Read, Self->prvFile, &read);
            if (read.Result != bodylen) {
               error = ObjectError(Self, ERH_Init, ERR_Read);
               goto exit;
            }
         }
      }
      else {
         seek.Position = POS_CURRENT;
         seek.Offset   = ChunkSize;
         if (Action(AC_Seek, Self->prvFile, &seek) != ERR_Okay) {
            ObjectError(Self, ERH_Init, ERR_Seek);
            goto exit;
         }
      }
   }

   if (BODY) {
      /* Query the Picture and then initialise the Bitmap before
      ** unpacking the file.
      */

      if (Action(AC_Query, Self, NULL) IS ERR_Okay) {
         Action(AC_Init, Self->Bitmap, NULL);
         error = UnpackPicture(Self, BMHD, Self->prvFile, CAMG, BODY, bodylen);
      }
      else DPrintF("@Picture:","Could not query picture information.");
   }
   else DPrintF("@Picture:","Picture has no BODY chunk.");

   /*** Check if a palette has been placed for this picture ***/

   if ((Self->Bitmap->AmtColours <= 256) AND (Self->Bitmap->Palette IS NULL)) {
      DPrintF("@Picture:","Warning - I could not obtain a palette from the file.");
   }

   if (Self->Flags & PCF_CACHE) Action(MT_CacheBitmap, Self, NULL);

exit:
   if (Buffer) FreeMemory(Buffer);
   if (BMHD)   FreeMemory(BMHD);
   if (BODY)   FreeMemory(BODY);
   Action(AC_Free, Self->prvFile, NULL);
   Self->prvFile = NULL;
   StepBack();
   return(error);
}

/*****************************************************************************

###PRIVATE###
Name:  CacheBitmap
Short: Adds the picture's bitmap to the cache.
###END###

*****************************************************************************/

static ERROR PIC_CacheBitmap(struct Picture *Self, APTR Void)
{
   struct FileInfo info;
   STRING location;
   WORD i, j;

   DPrintF("~5CacheBitmap()","[Picture:%d]", Self->Head.UniqueID);

   if ((GetField(Self, FID_Location, FT_STRING, &location) != ERR_Okay) OR (!location)) {
      StepBack();
      return(ERR_FieldNotSet);
   }

   if (glMemUsed > CACHE_SIZELIMIT) {
      /* If the cache has reached its size limit, scan for any unused bitmaps
      ** and free them so that we can bring the cache under control.
      */

      for (i=0; i < glCacheSize; i++) {
         if ((glCache[i].BitmapID) AND (glCache[i].OpenCount IS 0)) {
            DPrintF("CacheBitmap()","[Picture] Freeing cached bitmap #%d.", glCache[i].BitmapID);
            ActionMsg(AC_Free, glCache[i].BitmapID, NULL);
            if (glCache[i].MaskID) ActionMsg(AC_Free, glCache[i].MaskID, NULL);
            glMemUsed -= glCache[i].Size;
            ClearMemory(glCache + i, sizeof(glCache[i]));
         }
      }
   }

   /*** Find a place in the cache array ***/

   if (!glCache) {
      if (AllocPrivateMemory(sizeof(struct PictureCache) * 50, MEM_UNTRACKED, (void **)&glCache) IS ERR_Okay) {
         glCacheSize = 50;
      }
      else {
         Self->Flags &= ~PCF_CACHE;
         return(StepObjectError(Self, ERH_Init, ERR_AllocMemory));
      }
   }

   for (i=0; i < glCacheSize; i++) {
      if (glCache[i].BitmapID IS NULL) break;
   }

   if (i >= glCacheSize) {
      if (ReallocMemory(glCache, sizeof(struct PictureCache) * (glCacheSize + 10), &glCache, NULL) != ERR_Okay) {
         Self->Flags &= ~PCF_CACHE;
         return(StepObjectError(Self, ERH_Init, ERR_ReallocMemory));
      }
   }

   /* Track bitmaps back to our module so that we don't run into trouble
   ** during the shutdown process.  Note well that this is the correct way
   ** to do resource tracking in our case (using NF_UNTRACKED to allocate
   ** our bitmap object would cause all kinds of problems during shutdown).
   */

   if (SetOwner(Self->Bitmap, PictureModule) != ERR_Okay) {
      DPrintF("@CacheBitmap:","Failed to track the picture back to our module.");
      StepBack();
      return(ERR_Failed);
   }

   /*** Add the bitmap to cache ***/

   if (GetFileInfo(location, &info, sizeof(info)) IS ERR_Okay) {
      glCache[i].FileTime = info.TimeStamp;
      glCache[i].FileSize = info.Size;
   }
   else DPrintF("@CacheBitmap:","GetFileInfo() failed: %s", location);

   glCache[i].BitmapID  = Self->prvBitmapID;
   glCache[i].MaskID    = Self->prvMaskID;
   glCache[i].OpenCount = 1;
   glCache[i].Time      = SystemTime();
   glCache[i].Size      = Self->Bitmap->Size;
   for (j=0; (location[j]) AND (j < sizeof(glCache[i].Location)-1); j++) glCache[i].Location[j] = location[j];
   glCache[i].Location[j] = 0;

   glMemUsed += glCache[i].Size;

   Self->Flags |= PCF_CACHE;
   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Picture: Free
*/

static ERROR PIC_Free(struct Picture *Self, APTR Void)
{
   WORD i;

   if (Self->prvFile) {
      Action(AC_Free, Self->prvFile, NULL);
      Self->prvFile = NULL;
   }

   if (Self->prvLocationMID) {
      if (Self->prvLocation) {
         ReleaseMemory(Self->prvLocation);
         Self->prvLocation = NULL;
      }
      FreeMemoryID(Self->prvLocationMID);
      Self->prvLocationMID = NULL;
   }

   if (Self->prvDescriptionMID) {
      if (Self->prvDescription) {
         ReleaseMemory(Self->prvDescription);
         Self->prvDescription = NULL;
      }
      FreeMemoryID(Self->prvDescriptionMID);
      Self->prvDescriptionMID = NULL;
   }

   if (Self->prvDisclaimerMID) {
      if (Self->prvDisclaimer) {
         ReleaseMemory(Self->prvDisclaimer);
         Self->prvDisclaimer = NULL;
      }
      FreeMemoryID(Self->prvDisclaimerMID);
      Self->prvDisclaimerMID = NULL;
   }

   for (i=0; i < glCacheSize; i++) {
      if (glCache[i].BitmapID IS Self->prvBitmapID) {
         glCache[i].OpenCount--;

         /*** Release our locks but do not free the cached bitmap ***/
         if (Self->Bitmap) ReleaseObject(Self->Bitmap);
         if (Self->Mask)   ReleaseObject(Self->Mask);
         Self->prvBitmapID = NULL;
         Self->Bitmap      = NULL;
         Self->prvMaskID   = NULL;
         Self->Mask        = NULL;
         break;
      }
   }

   if (Self->prvBitmapID) {
      if (Self->Bitmap) {
         Action(AC_Free, Self->Bitmap, NULL);
         ReleaseObject(Self->Bitmap);
         Self->Bitmap = NULL;
      }
      else ActionMsg(AC_Free, Self->prvBitmapID, NULL);
      Self->prvBitmapID = NULL;
   }

   if (Self->prvMaskID) {
      if (Self->Mask) {
         Action(AC_Free, Self->Mask, NULL);
         ReleaseObject(Self->Mask);
         Self->Mask = NULL;
      }
      else ActionMsg(AC_Free, Self->prvMaskID, NULL);
      Self->prvMaskID = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Init
Short: Prepares picture objects for usage.

###DESCRIPTION###
<p>Objects that belong to the Picture class can be initialised in two
possible ways.  If you have not set the <a href="#Location">Location</a>
field or have chosen to use the NEW <a href="#Flag">Flag</a>, the
initialisation routine will create a <a href="#Bitmap">Bitmap</a> area that
contains no image data.  This allows you to fill the picture with your own
image data and save it using the SaveImage or SaveToObject actions.  You must
set the Bitmap's Width, Height and colour specifications at a minimum, or
the initialisation process will fail.</p>

<p>If you have set the <a href="#Location">Location</a> field and avoided
the NEW flag, the initialisation process will analyse the file location to
determine whether or not the data is in fact a valid image file.  If the file
does not match up with a registered data format, an error code of
ERR_NoSupport is returned.  You will need to use the Activate or Query
actions to load or find out more information about the image format.</p>

###END###

*****************************************************************************/

static ERROR PIC_Init(struct Picture *Self, APTR Void)
{
   struct Bitmap *bitmap;
   struct acRead read;
   struct FileInfo info;
   STRING location;
   WORD i;

   if ((GetField(Self, FID_Location, FT_STRING, &location) != ERR_Okay) OR (!location)) {
      return(ERR_FieldNotSet);
   }

   if ((!location) OR (Self->Flags & PCF_NEW)) {
      /* If no location has been specified, assume that the picture is being
      ** created from scratch (e.g. to save an image to disk).  The
      ** programmer is required to specify the dimensions and colours of the
      ** Bitmap so that we can initialise it.
      */

      if (!Self->Bitmap->Width) Self->Bitmap->Width = Self->DisplayWidth;
      if (!Self->Bitmap->Height) Self->Bitmap->Height = Self->DisplayHeight;

      if ((Self->Bitmap->Width) AND (Self->Bitmap->Height)) {
         if (Action(AC_Init, Self->Bitmap, NULL) IS ERR_Okay) {

            if (Self->Flags & PCF_ALPHA) {
               if (NewObject(ID_BITMAP, Self->Head.ObjectFlags|NF_CHILD, &Self->Mask, &Self->prvMaskID) IS ERR_Okay) {
                  Self->Mask->Width  = Self->Bitmap->Width;
                  Self->Mask->Height = Self->Bitmap->Height;
                  Self->Mask->Flags |= PCF_MASK;
                  Self->Mask->BitsPerPixel = 8;
                  if (Action(AC_Init, Self->Mask, NULL) IS ERR_Okay) Self->Flags |= PCF_MASK;
                  else return(ObjectError(Self, ERH_Init, ERR_Init));
               }
               else return(ObjectError(Self, ERH_Init, ERR_NewObject));
            }
            else if (Self->Flags & PCF_MASK) {
               if (NewObject(ID_BITMAP, Self->Head.ObjectFlags|NF_CHILD, &Self->Mask, &Self->prvMaskID) IS ERR_Okay) {
                  Self->Mask->Width  = Self->Bitmap->Width;
                  Self->Mask->Height = Self->Bitmap->Height;
                  Self->Mask->Flags |= PCF_MASK;
                  Self->Mask->BitsPerPixel = 1;
                  if (Action(AC_Init, Self->Mask, NULL) IS ERR_Okay) Self->Flags |= PCF_MASK;
                  else return(ObjectError(Self, ERH_Init, ERR_Init));
               }
               else return(ObjectError(Self, ERH_Init, ERR_NewObject));
            }

            return(ERR_Okay);
         }
         else return(ObjectError(Self, ERH_Init, ERR_Init));
      }
      else return(ObjectError(Self, ERH_Init, ERR_FieldNotSet));
   }
   else {

      /*** Use a cached bitmap if possible ***/

      if ((Self->Flags & PCF_CACHE) AND (glCache) AND (Self->Head.UniqueID > 0)) {
         for (i=0; i < glCacheSize; i++) {
            if (glCache[i].BitmapID) {
               if (StrCompare(glCache[i].Location, location, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  /* We have a filename match.  Check the original file location's size and date-time
                  ** to see if the original file has changed before we commit to using
                  ** the cached version.
                  */

                  if (GetFileInfo(location, &info, sizeof(info)) != ERR_Okay) continue;

                  if (info.TimeStamp != glCache[i].FileTime) continue;
                  if (info.Size != glCache[i].FileSize) continue;

                  if (AccessObject(glCache[i].BitmapID, 5000, (OBJECTPTR *)&bitmap) IS ERR_Okay) {
                     if (Self->Bitmap) {
                        /*** Free the existing bitmap object for our picture ***/
                        Action(AC_Free, Self->Bitmap, NULL);
                        ReleaseObject(Self->Bitmap);
                     }

                     Self->Bitmap = bitmap;
                     Self->prvBitmapID = glCache[i].BitmapID;
                     if (glCache[i].MaskID) {
                        AccessObject(glCache[i].MaskID, 5000, (OBJECTPTR *)&Self->Mask);
                        Self->prvMaskID = glCache[i].MaskID;
                     }

                     glCache[i].OpenCount++;
                     glCache[i].Time = SystemTime();

                     Self->DisplayWidth  = Self->Bitmap->Width;
                     Self->DisplayHeight = Self->Bitmap->Height;

                     return(ERR_Okay);
                  }
                  break;
               }
            }
         }
      }

      /* Test the given location to see if it matches
      ** our supported file format.
      */

      if (CreateObject(ID_FILE, NULL, &Self->prvFile, NULL,
            FID_Location|TSTRING, location,
            FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
            TAGEND) IS ERR_Okay) {

         read.Buffer = Self->prvHeader;
         read.Length = sizeof(Self->prvHeader);
         Action(AC_Read, Self->prvFile, &read);

         if ((StrCompare(Self->prvHeader, "FORM", 4, STR_CASE) IS ERR_Okay) AND
             (StrCompare(Self->prvHeader + 8, "ILBM", 4, STR_CASE) IS ERR_Okay)) {
            if (!(Self->Flags & PCF_QUERY)) Action(AC_Activate, Self, NULL);
            return(ERR_Okay);
         }
         else return(ERR_NoSupport);
      }
      else return(ObjectError(Self, ERH_Init, ERR_CreateObject));
   }
}

/*****************************************************************************
** Picture: NewObject
*/

static ERROR PIC_NewObject(struct Picture *Self, APTR Void)
{
   Self->Quality = 80; /* 80% quality rating when saving */

   if (NewObject(ID_BITMAP, Self->Head.ObjectFlags|NF_CHILD, &Self->Bitmap, &Self->prvBitmapID) IS ERR_Okay) {
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_NewObject, ERR_NewObject));
}

/*****************************************************************************

###ACTION###
Name:  Query
Short: Reads information from a picture's file source and fills the object structure.

###DESCRIPTION###
<p>Calling this action will fill out the fields of a picture object according
to the image referred to in the <a href="#Location">Location</a> field. This
includes information on the width, height, palette values, screen dimensions
and so on.</p>

<p>Fields that contain preset values will not be over-written by this
action.</p>

###END###

*****************************************************************************/

static ERROR PIC_Query(struct Picture *Self, APTR Void)
{
   struct BMHD   *BMHD;
   struct Bitmap *Bitmap;
   struct acSeek seek;
   struct acRead read;
   LONG error, CAMG, *CMAP, totalcolours, i, FileSize, ChunkSize, Position;
   UBYTE *Buffer;
   UBYTE *SrcPalette;

   BMHD     = 0;
   CAMG     = 0;
   CMAP     = 0;
   totalcolours = 0;
   Buffer   = 0;
   error    = ERR_Failed;

   DPrintF("~Query()","Picture: $%x", Self);

   Bitmap = Self->Bitmap;

   if (AllocMemory(BUFFERSIZE, MEM_NOCLEAR, (APTR *)&Buffer, NULL) != ERR_Okay) goto exit;

   GetField(Self->prvFile, FID_Size, FT_LONG, &FileSize);

   ActionTags(AC_Seek, Self->prvFile, 12.0, POS_START);
   Position = 12;
   while (Position < FileSize) {
      read.Buffer = Buffer;
      read.Length = 8;
      Action(AC_Read, Self->prvFile, &read);
      if (read.Result != 8) { /* [CODE] [Size] */
         DPrintF("!Query:","I could not read the next chunk.");
         break;
      }

      ChunkSize = rd_long(Buffer+4);
      if (ChunkSize & 0x00000001) ChunkSize++;

      if (StrCompare(Buffer,"CMAP", 4, STR_CASE) IS ERR_Okay) {
         DPrintF("3Query:","IFF CMAP header found.");
         if (ChunkSize > BUFFERSIZE) goto exit;

         totalcolours = ChunkSize/3;
         if (AllocMemory(ChunkSize, MEM_NOCLEAR, (APTR *)&CMAP, NULL) IS ERR_Okay) {
            read.Buffer = (APTR)CMAP;
            read.Length = ChunkSize;
            Action(AC_Read, Self->prvFile, &read);
         }
      }
      else if (StrCompare(Buffer,"BMHD", 4, STR_CASE) IS ERR_Okay) {
         DPrintF("3Query:","IFF BMHD header found.");
         if (ChunkSize > BUFFERSIZE) goto exit;
         if (AllocMemory(ChunkSize, MEM_NOCLEAR, (APTR *)&BMHD, NULL) IS ERR_Okay) {
            struct mtReadFileWord Args;
            Action(MT_ReadFileWord, Self->prvFile, &Args); BMHD->Width  = Args.Result;
            Action(MT_ReadFileWord, Self->prvFile, &Args); BMHD->Height = Args.Result;
            Action(MT_ReadFileWord, Self->prvFile, &Args); BMHD->X = Args.Result;
            Action(MT_ReadFileWord, Self->prvFile, &Args); BMHD->Y = Args.Result;
            SetRead(read, &BMHD->Depth, 1);       Action(AC_Read, Self->prvFile, &read);
            SetRead(read, &BMHD->Mask, 1);        Action(AC_Read, Self->prvFile, &read);
            SetRead(read, &BMHD->Pack, 1);        Action(AC_Read, Self->prvFile, &read);
            SetRead(read, &BMHD->PAD, 1);         Action(AC_Read, Self->prvFile, &read);
            Action(MT_ReadFileWord, Self->prvFile, &Args); BMHD->TColor = Args.Result;
            SetRead(read, &BMHD->XAspect, 1);     Action(AC_Read, Self->prvFile, &read);
            SetRead(read, &BMHD->YAspect, 1);     Action(AC_Read, Self->prvFile, &read);
            Action(MT_ReadFileWord, Self->prvFile, &Args); BMHD->ScrWidth  = Args.Result;
            Action(MT_ReadFileWord, Self->prvFile, &Args); BMHD->ScrHeight = Args.Result;
            DPrintF("Query:","[IFF] File dimensions are %dw,%dh,%dbpp, packed: %d", BMHD->Width, BMHD->Height, BMHD->Depth, BMHD->Pack);
         }
         else {
            error = ObjectError(Self, ERH_Query, ERR_Memory);
            goto exit;
         }
      }
      else if (StrCompare(Buffer, "CAMG", 4, STR_CASE) IS ERR_Okay) {
         DPrintF("3Query:","IFF CAMG header found.");
         if (ChunkSize > BUFFERSIZE) goto exit;
         SetRead(read, Buffer, ChunkSize);
         Action(AC_Read, Self->prvFile, &read);
         if (read.Result IS ChunkSize) {
            CAMG = rd_long(Buffer);
         }
         else {
            error = ObjectError(Self, ERH_Query, ERR_Read);
            goto exit;
         }
      }
      else {
         DPrintF("3Query:","Ignoring chunk %.4s, skipping %ld bytes.", Buffer, ChunkSize);

         seek.Position = POS_CURRENT;
         seek.Offset   = ChunkSize;
         if (Action(AC_Seek, Self->prvFile, &seek) != ERR_Okay) { /* Skip to the next chunk */
             ObjectError(Self, ERH_Query, ERR_Seek);
             goto exit;
         }
      }
      GetField(Self->prvFile, FID_Position, FT_LONG, &Position);
   }

   DPrintF("Query:","File has been read, processing data...");

   if (BMHD) {
      if (!Bitmap->AmtColours)     Bitmap->AmtColours   = totalcolours;
      if (!Bitmap->Height)         Bitmap->Height       = BMHD->Height;
      if (!Bitmap->BitsPerPixel)   Bitmap->BitsPerPixel = 8;
      if (!Bitmap->Type)           Bitmap->Type         = BMP_CHUNKY;
      if (!Self->DisplayWidth)  Self->DisplayWidth  = BMHD->ScrWidth;
      if (!Self->DisplayHeight) Self->DisplayHeight = BMHD->ScrHeight;
      if (!Bitmap->Width) {
         Bitmap->Width     = BMHD->Width;
         Bitmap->ByteWidth = 0;         /* Let the Init(Bitmap) calculate it */
      }
   }

   /*** Generate Palette ***/

   if ((Bitmap->Palette) AND (CMAP)) {
      SrcPalette = (BYTE *)CMAP;
      for (i=0; i < totalcolours; i++) {
         Bitmap->Palette->Col[i].Red   = SrcPalette[0];
         Bitmap->Palette->Col[i].Green = SrcPalette[1];
         Bitmap->Palette->Col[i].Blue  = SrcPalette[2];
         SrcPalette  += 3;
      }
   }

   /*** Re-query the Picture Bitmap ***/

   if (Action(AC_Query, Bitmap, NULL) != ERR_Okay) {
      error = ObjectError(Self, ERH_Query, ERR_Query);
      goto exit;
   }

   error = ERR_Okay;

exit:
   if (Buffer) FreeMemory(Buffer);
   if (BMHD)   FreeMemory(BMHD);
   if (CMAP)   FreeMemory(CMAP);
   StepBack();
   return(error);
}

/*****************************************************************************

###ACTION###
Name:  Read
Short: Reads raw image data from a Picture object.
###END###

*****************************************************************************/

static ERROR PIC_Read(struct Picture *Self, struct acRead *Args)
{
   return(Action(AC_Read, Self->Bitmap, Args));
}

/*****************************************************************************
** Picture: ReleaseObject
*/

static ERROR PIC_ReleaseObject(struct Picture *Self, APTR Void)
{
   if (Self->prvLocation)    { ReleaseMemory(Self->prvLocation);    Self->prvLocation    = NULL; }
   if (Self->prvDescription) { ReleaseMemory(Self->prvDescription); Self->prvDescription = NULL; }
   if (Self->prvDisclaimer)  { ReleaseMemory(Self->prvDisclaimer);  Self->prvDisclaimer  = NULL; }
   if (Self->Bitmap) { ReleaseObject(Self->Bitmap); Self->Bitmap = NULL; }
   if (Self->Mask)   { ReleaseObject(Self->Mask);   Self->Mask = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  SaveImage
Short: Saves a Picture image to a data object.
###END###

*****************************************************************************/

static ERROR PIC_SaveImage(struct Picture *Self, struct acSaveImage *Args)
{
   return(ActionTags(AC_SaveImage, Self->Bitmap, Args->DestinationID, NULL));
}

/*****************************************************************************

###ACTION###
Name:  SaveToObject
Short: Saves a Picture image to a data object.
###END###

*****************************************************************************/

static ERROR PIC_SaveToObject(struct Picture *Self, struct acSaveToObject *Args)
{
   struct SysClass *Class;
   ERROR (**Routine)(OBJECTPTR, APTR);

   if ((Args->ClassID) AND (Args->ClassID != ID_PICTURE)) {
      Class = (struct SysClass *)FindClass(Args->ClassID, NULL);
      if ((GetField(Class, FID_ActionTable, FT_POINTER, &Routine) IS ERR_Okay) AND (Routine)) {
         if (Routine[AC_SaveToObject]) {
            return(Routine[AC_SaveToObject]((OBJECTPTR)Self, Args));
         }
         else return(ObjectError(Self, ERH_SaveToObject, ERR_NoSupport));
      }
      else return(ObjectError(Self, ERH_SaveToObject, ERR_GetField));
   }
   else {
      return(ActionTags(AC_SaveImage, Self, Args->DestinationID, Args->ClassID));
   }
}

/*****************************************************************************

###ACTION###
Name:  Seek
Short: Seeks to a new read/write position within a Picture object.
###END###

*****************************************************************************/

static ERROR PIC_Seek(struct Picture *Self, struct acSeek *Args)
{
   return(Action(AC_Seek, Self->Bitmap, Args));
}

/*****************************************************************************

###ACTION###
Name:  Write
Short: Writes raw image data to a Picture object.
###END###

*****************************************************************************/

static ERROR PIC_Write(struct Picture *Self, struct acWrite *Args)
{
   return(Action(AC_Write, Self->Bitmap, Args));
}

/*****************************************************************************

###FIELD###
Name:   Author
Short:  The name of the person or company that created the image.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The name of the person or organisation that created a Picture's image is
normally specified here.  This provides a means for the original author to
claim credit for the image.</p>

###SEE ALSO###
Field: Copyright

###END###

*****************************************************************************/

static ERROR GET_Author(struct Picture *Self, STRING *Value)
{
   *Value = Self->prvAuthor;
   return(ERR_Okay);
}

static ERROR SET_Author(struct Picture *Self, STRING Value)
{
   LONG i;
   for (i=0; (Value[i]) AND (i < sizeof(Self->prvAuthor)-1); i++) Self->prvAuthor[i] = Value[i];
   Self->prvAuthor[i] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Bitmap
Short:  Represents a Picture's image data.
Type:   struct Bitmap *
Status: Read

###DESCRIPTION###
<p>The details of a Picture's graphical image and data are defined in its
associated Bitmap object.  It contains information on the image dimensions
and palette for example.  When loading a Picture, you can place certain
constraints on the image by presetting Bitmap fields such as the Width and
Height (this will have the effect of clipping or resizing the source image).
The Palette can also be preset if you want to remap the source image to a
specific set of colour values.</p>

<p>Please refer to the <a href="bitmap.html">Bitmap</a> class for more
details on the structure of the Bitmap object.</p>

###SEE ALSO###
Class: Bitmap

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Copyright
Short:  Copyright details of an image.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>Copyright details related to an image may be specified here.  The
copyright should be short and to the point, for example &quot;Copyright H.R.
Giger (c) 1992.  All rights reserved.&quot;</p>

###SEE ALSO###
Field: Author

###END###

*****************************************************************************/

static ERROR GET_Copyright(struct Picture *Self, STRING *Value)
{
   *Value = Self->prvCopyright;
   return(ERR_Okay);
}

static ERROR SET_Copyright(struct Picture *Self, STRING Value)
{
   LONG i;
   for (i=0; (Value[i]) AND (i < sizeof(Self->prvCopyright)-1); i++) Self->prvCopyright[i] = Value[i];
   Self->prvCopyright[i] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Description
Short:  Long description for an image.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>A long description for an image may be entered in this field.  There is
no strict limit on the length of the description.  It is recommended that
this field be used to describe what the image is of.</p>

###END###

*****************************************************************************/

static ERROR GET_Description(struct Picture *Self, STRING *Value)
{
   if (Self->prvDescription) {
      *Value = Self->prvDescription;
      return(ERR_Okay);
   }
   else if (!Self->prvDescriptionMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->prvDescriptionMID, MEM_READ, (void **)&Self->prvDescription) IS ERR_Okay) {
      *Value = Self->prvDescription;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Description(struct Picture *Self, STRING Value)
{
   LONG i;

   if (Self->prvDescription)    { ReleaseMemory(Self->prvDescription);   Self->prvDescription = NULL; }
   if (Self->prvDescriptionMID) { FreeMemoryID(Self->prvDescriptionMID); Self->prvDescriptionMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->prvDescription, &Self->prvDescriptionMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->prvDescription[i] = Value[i];
         Self->prvDescription[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Disclaimer
Short:  The disclaimer associated with an image.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>If it is necessary to associate a disclaimer with an image, the legal
text may be entered in this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Disclaimer(struct Picture *Self, STRING *Value)
{
   if (Self->prvDisclaimer) {
      *Value = Self->prvDisclaimer;
      return(ERR_Okay);
   }
   else if (!Self->prvDisclaimerMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->prvDisclaimerMID, MEM_READ, (void **)&Self->prvDisclaimer) IS ERR_Okay) {
      *Value = Self->prvDisclaimer;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Disclaimer(struct Picture *Self, STRING Value)
{
   LONG i;

   if (Self->prvDisclaimer)    { ReleaseMemory(Self->prvDisclaimer);   Self->prvDisclaimer = NULL; }
   if (Self->prvDisclaimerMID) { FreeMemoryID(Self->prvDisclaimerMID); Self->prvDisclaimerMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->prvDisclaimer, &Self->prvDisclaimerMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->prvDisclaimer[i] = Value[i];
         Self->prvDisclaimer[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELDS###
Names:  DisplayHeight
Short:  The preferred height to use when displaying the Picture.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field defines the preferred height of the display when viewing the
Picture image.  If a picture file does not contain a suitable value to be
placed in this field, it will be initialised to a value based on the
Picture's resolution and <a href="#Bitmap">Bitmap</a> Height.</p>

###SEE ALSO###
Field: DisplayWidth

###END###

*****************************************************************************/

/*****************************************************************************

###FIELDS###
Names:  DisplayWidth
Short:  The preferred width to use when displaying the Picture.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field defines the preferred width of the display when viewing the
Picture image.  If a picture file does not contain a suitable value to be
placed in this field, it will be initialised to a value based on the
Picture's resolution and <a href="#Bitmap">Bitmap</a> Width.</p>

###SEE ALSO###
Field: DisplayHeight

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Prefix: PCF
Short:  Optional initialisation flags.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>You can specify flags in this field that will affect the way that image
files are loaded.  Valid flags are:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">ALPHA</td><td>If a picture supports an alpha channel, this field will be set (if you will be saving a Picture with an alpha channel, set this flag prior to initialisation and build the channel using the object provided in the Mask field).</td></tr>
<tr><td width="10%">NEW</td><td>If a picture is being created from scratch (i.e. you will not be loading any image data), set the NEW flag so that the initialisation process will not attempt to load any picture data.</td></tr>
<tr><td width="10%">NOCOMPARE&nbsp;&nbsp;</td><td>When loading images of 256 colours or less, the picture code will automatically compare the image palette to the Picture's Bitmap Palette. If differences are found then the code will remap the picture by default. However, you may want to allow for differences in the palette, in which case you can set PCF_NOCOMPARE to stop palette comparisons.  Please note that if you want to acquire the source palette prior to loading a picture, use the Query action prior to Activate.</td></tr>
<tr><td width="10%">CACHE</td><td>Use the CACHE option to put the loaded bitmap image into a global cache.  Any further picture objects that load the referenced image file will use the cached bitmap rather than re-loading the image data.</td></tr>
<tr><td width="10%">32BITALPHA</td><td>If the picture source includes an alpha channel mask, it will normally be separated out and stored in the Mask field.  If you set the 32BITALPHA option before loading, the image will be forced into 32-bit graphics mode with alpha channel.  The mask information will be stored in the alpha byte and the Mask field will not be set.</td></tr>
<tr><td width="10%">QUERY</td><td>This option may be set prior to loading an image in order to prevent the loading and decompression of the image data.  Only the picture's information header will be processed.</td></tr>
<tr><td width="10%">BACKGROUND</td><td>This flag is set automatically if a loaded image carries a background colour.  The colour is indicated in the bitmap's ColourRGB field.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Header
Short:  Contains the first 32 bytes of data in a picture's file header.
Type:   POINTER
Status: Get

###DESCRIPTION###
<p>The Header field is a pointer to a 32 byte buffer that contains the first
32 bytes of information read from a picture file on initialisation.  This
special field is considered to be helpful only to developers writing add on
components for the picture class.</p>

<p>The buffer that is referred to by the Header field is not populated until
the Init action is called on the picture object.</p>

###END###

*****************************************************************************/

static ERROR GET_Header(struct Picture *Self, APTR *Value)
{
   *Value = Self->prvHeader;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Location
Short:  The location of a Picture's image data.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>This compulsory field must refer to a file that contains the image data
that you want to load.</p>

###END###

*****************************************************************************/

static ERROR GET_Location(struct Picture *Self, STRING *Value)
{
   if (Self->prvLocation) {
      *Value = Self->prvLocation;
      return(ERR_Okay);
   }
   else if (!Self->prvLocationMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->prvLocationMID, MEM_READ, (void **)&Self->prvLocation) IS ERR_Okay) {
      *Value = Self->prvLocation;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Location(struct Picture *Self, STRING Value)
{
   LONG i;

   if (Self->prvLocation)    { ReleaseMemory(Self->prvLocation);   Self->prvLocation = NULL; }
   if (Self->prvLocationMID) { FreeMemoryID(Self->prvLocationMID); Self->prvLocationMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->prvLocation, &Self->prvLocationMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->prvLocation[i] = Value[i];
         Self->prvLocation[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Software
Short:  The name of the application that was used to draw the image.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The Software field declares the name of the application that was used to
draw or render the original image.</p>

###END###

*****************************************************************************/

static ERROR GET_Software(struct Picture *Self, STRING *Value)
{
   *Value = Self->prvSoftware;
   return(ERR_Okay);
}

static ERROR SET_Software(struct Picture *Self, STRING Value)
{
   LONG i;
   for (i=0; (Value[i]) AND (i < sizeof(Self->prvSoftware)-1); i++) Self->prvSoftware[i] = Value[i];
   Self->prvSoftware[i] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Title
Short:  The title of the image.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>If the author of a picture has assigned a title to it, then it must be
declared through the Title field.</p>

###END###

*****************************************************************************/

static ERROR GET_Title(struct Picture *Self, STRING *Value)
{
   *Value = Self->prvTitle;
   return(ERR_Okay);
}

static ERROR SET_Title(struct Picture *Self, STRING Value)
{
   LONG i;
   for (i=0; (Value[i]) AND (i < sizeof(Self->prvTitle)-1); i++) Self->prvTitle[i] = Value[i];
   Self->prvTitle[i] = 0;
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: UnpackPicture()
**
** Unpacks the BODY data to the Picture->Bitmap.  If the palettes or amount
** of colours do not match, the picture will be automatically remapped.
*/

static LONG UnpackPicture(struct Picture *Self, struct BMHD *BMHD,
               OBJECTPTR File, LONG CAMG, BYTE *BODY, LONG BodyLength)
{
   struct mtDrawRGBPixel pixel;
   struct mtReadRGBPixel read;
   struct Bitmap *Bitmap, *ILBMBitmap;
   LONG   ecode;
   WORD   MaxWidth, y, j, ydest, XRemainder, YRemainder, Domain, Height;
   BYTE   *Dest, *BODYEnd;
   ERROR (*QDrawRGBPixel)(struct Bitmap *, struct mtDrawRGBPixel *);
   ERROR (*QReadRGBPixel)(struct Bitmap *, struct mtReadRGBPixel *);

   ILBMBitmap = NULL;
   ecode      = ERR_Failed;

   if (Self->Flags & PCF_RESIZE) {
      DPrintF("UnpackPicture:","Note that the image will need to be resized.");
   }

   Bitmap     = Self->Bitmap;
   YRemainder = 0;
   if ((Height = Bitmap->Height) > BMHD->Height) {
      Height = BMHD->Height;
   }

   if (ObtainMethod(Bitmap, MT_DrawRGBPixel, (void **)&QDrawRGBPixel) != ERR_Okay) return(ERR_Failed);
   if (ObtainMethod(Bitmap, MT_ReadRGBPixel, (void **)&QReadRGBPixel) != ERR_Okay) return(ERR_Failed);

   /*** Calculate the amount of colours in the ILBM BMHD source. ***/

   DPrintF("3UnpackPicture:","Allocating a single-line decompression bitmap.");

   if (NewPrivateObject(ID_BITMAP, NULL, &ILBMBitmap) IS ERR_Okay) {
      ILBMBitmap->Width        = BMHD->Width;
      ILBMBitmap->Height       = 1;
      ILBMBitmap->BitsPerPixel = BMHD->Depth;
      ILBMBitmap->Type         = BMP_ILBM;
      ILBMBitmap->Palette      = Self->Bitmap->Palette;
      ILBMBitmap->AmtColours   = 1<<ILBMBitmap->BitsPerPixel;

      if (ILBMBitmap->BitsPerPixel IS 24) ILBMBitmap->BytesPerPixel = 3;
      else if (ILBMBitmap->BitsPerPixel <= 8) ILBMBitmap->BytesPerPixel = 1;

      if (Action(AC_Init, ILBMBitmap, NULL) != ERR_Okay) goto exit;
   }
   else goto exit;

   /*** Calculate some initial variables for the loop ***/

   if (Bitmap->Width < ILBMBitmap->Width) MaxWidth = Bitmap->Width;
   else MaxWidth = ILBMBitmap->Width;

   ydest = NULL;

   /* Begin the loop now.  Unpack one row at a time to the ILBM Bitmap
   ** buffer, then copy the pixels from the buffer over to our destination.
   */

   if (CAMG & OSV_HAM) {
      DPrintF("!Activate:","[IFF] HAM IFF picture files are not supported.");
      ecode = ERR_BadData;
      goto exit;
   }

   BODYEnd = BODY + BodyLength;
   for (y=0; y < Height; y++) {
      if (BODY >= BODYEnd) {
         DPrintF("@Activate:","[IFF] Buffer overflow at line %d.", y);
         ecode = ERR_Okay;
         goto exit;
      }

      /*** Unpack the next line of body data to our one-line interleaved bitmap buffer ***/

      Dest = ILBMBitmap->Data;
      for (j=0; j < ILBMBitmap->BitsPerPixel; j++) {
         BODY  = UnpackPlane(BMHD, File, ILBMBitmap, Dest, BODY);
         Dest += ILBMBitmap->ByteWidth;
      }

      /* Write the data out to the destination Bitmap.  If resizing,
      ** we use the standard DrawPixel() and ReadPixel() routines.  Although
      ** we could improve the speed in some parts, it is not too crucial.
      */

      if ((Self->Flags & PCF_RESIZEX) AND (Bitmap->Width != ILBMBitmap->Width)) {
         if (ILBMBitmap->Width > Bitmap->Width) { /*** Shrink Row ***/
            XRemainder   = 0;
            Domain       = ILBMBitmap->Width - Bitmap->Width;
            read.XCoord  = 0;
            read.YCoord  = 0;
            pixel.YCoord = ydest;
            for (pixel.XCoord=0; pixel.XCoord < Bitmap->Width; pixel.XCoord++) {
               XRemainder += Domain;
               while (XRemainder >= Bitmap->Width) {
                  XRemainder -= Bitmap->Width;
                  read.XCoord++;
               }
               QReadRGBPixel(ILBMBitmap, &read);
               pixel.Red   = read.Red;
               pixel.Green = read.Green;
               pixel.Blue  = read.Blue;
               QDrawRGBPixel(Bitmap, &pixel);
               read.XCoord++;
            }
         }
         else { /*** Expand Row ***/
            XRemainder   = 0;
            Domain       = Bitmap->Width - ILBMBitmap->Width;
            read.YCoord  = 0;
            pixel.XCoord = 0;
            pixel.YCoord = ydest;

            for (read.XCoord=0; read.XCoord < ILBMBitmap->Width; read.XCoord++) {
               QReadRGBPixel(ILBMBitmap, &read);
               pixel.Red   = read.Red;
               pixel.Green = read.Green;
               pixel.Blue  = read.Blue;
               QDrawRGBPixel(Bitmap, &pixel);
               pixel.XCoord++;
               XRemainder += Domain;
               while (XRemainder >= ILBMBitmap->Width) {
                  QDrawRGBPixel(Bitmap, &pixel);
                  pixel.XCoord++;
                  XRemainder -= ILBMBitmap->Width;
               }
            }
         }
      }
      else {
         read.YCoord = 0;
         pixel.YCoord = ydest;
         for (pixel.XCoord=0; pixel.XCoord < MaxWidth; pixel.XCoord++) {
             read.XCoord = pixel.XCoord;
             QReadRGBPixel(ILBMBitmap, &read);
             pixel.Red   = read.Red;
             pixel.Green = read.Green;
             pixel.Blue  = read.Blue;
             QDrawRGBPixel(Bitmap, &pixel);
         }
      }

      /*** Image Resizing Y Axis ***/

      if (Self->Flags & PCF_RESIZEY) {
         if (Bitmap->Height < BMHD->Height) { /*** Shrink Down ***/
            Domain      = BMHD->Height - Bitmap->Height;
            YRemainder += Domain;
            while (YRemainder >= Bitmap->Height) {
               YRemainder -= Bitmap->Height;
               BODY = SkipLine(BMHD, BODY, File, ILBMBitmap);
            }
         }
         else if (Bitmap->Height > BMHD->Height) { /*** Expand ***/
            Domain      = Bitmap->Height - BMHD->Height;
            YRemainder += Domain;
            while (YRemainder >= BMHD->Height) {
               YRemainder -= BMHD->Height;
               ydest++;
               ActionTags(MT_BlitArea, Bitmap, Bitmap, NULL, 0, ydest-1,
                 Bitmap->Width, 1, 0, ydest);
            }
         }
      }
      ydest++;
   }

   ecode = ERR_Okay;

exit:
   if (ILBMBitmap) Action(AC_Free, ILBMBitmap, NULL);
   return(ecode);
}

/*****************************************************************************
** Internal: SkipLine()
** Short:    Skips a complete line of Buffer data.
*/

static BYTE * SkipLine(struct BMHD *BMHD, BYTE *Buffer, OBJECTPTR File,
                struct Bitmap *ILBMBitmap)
{
  WORD written;
  BYTE num, j;

  if (BMHD->Pack) {
     for (j=0; j < ILBMBitmap->BitsPerPixel; j++) {
        written = 0;
        while (written < ILBMBitmap->ByteWidth) {
           num = *Buffer++;

           if (num >= 0) {
              do {
                 Buffer++;
                 num--;
                 written++;
              } while (num >= 0);
           }
           else if (num != -128) {
              num = -num;
              Buffer++;

              do {
                 written++;
                 num--;
              } while (num >= 0);
           }
        }
     }
  }
  else for (j=0; j < ILBMBitmap->BitsPerPixel; j++) {
     for (num=0; num < ILBMBitmap->ByteWidth; num++) Buffer++;
  }

  return(Buffer);
}

/*****************************************************************************
** Internal: UnpackPlane()
** Short:    Unpacks one plane of BODY data to the ILBMBitmap object.
*/

static BYTE * UnpackPlane(struct BMHD *BMHD, OBJECTPTR File, struct Bitmap *ILBMBitmap,
                   BYTE *Dest, BYTE *Buffer)
{
  WORD written, i;
  BYTE num, col;

  if (BMHD->Pack) {
     written = 0;
     while (written < ILBMBitmap->ByteWidth) {
        num = *Buffer;
        Buffer++;

        if (num >= 0) {
           do {
              *Dest++ = *Buffer; Buffer++;
              num--;
              written++;
              if (written > ILBMBitmap->ByteWidth) return(Buffer);
           } while (num >= 0);
        }
        else if (num != -128) {
           col = *Buffer;
           Buffer++;
           do {
              *Dest++ = col;
              written++;
              num++;
              if (written > ILBMBitmap->ByteWidth) return(Buffer);
           } while (num <= 0);
        }
     }
  }
  else for (i=0; i < ILBMBitmap->ByteWidth; i++) Dest[i] = Buffer[i];

  return(Buffer);
}

