/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Uninstall
Version:   1.0
Status:    Stable
ID:        ID_UNINSTALL
Category:  Tool
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2002-2004.  All rights reserved.
Keywords:  uninstall, remove, delete, expunge, destroy, setup
Short:     Uninstalls software packages using XML-Uninstall files.

###DESCRIPTION###

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/files/all.h>
#include <pandora/data/config.h>
#include <pandora/data/xml.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_UNINSTALL 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_UNINSTALL, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2002-2004.  All rights reserved.",
   "January 2004",
   "Uninstall"
};

static struct KernelBase *KernelBase;
static struct StringsBase *StringsBase;
static struct FileSystemBase *FileSystemBase;
static OBJECTPTR UninstallClass = 0;
static OBJECTPTR StringsMod = 0, FileSystemMod = 0;

struct DeleteDir {
   struct DeleteDir *Next;
   UBYTE Path[];
};

static void DeleteConfigEntry(OBJECTPTR, STRING, STRING);

/*****************************************************************************
** Class definition.
*/

struct Uninstall {
   OBJECT_HEADER        /* Class header */
   OBJECTID OutputID;   /* Reference to the file navigation area */
   LONG Static;

   PRIVATE_FIELDS
   struct XMLTag *prvTag;
   OBJECTPTR prvXML;
   STRING prvLocation;
   struct DeleteDir *prvDirs;
};

static ERROR GET_Location(struct Uninstall *, STRING *);

static ERROR SET_Location(struct Uninstall *, STRING);

static struct FieldArray UninstallFields[] = {
   { "Output", 0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "Static", 0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   /*** Virtual fields ***/
   { "Location", 0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   { "Src",      0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   END_FIELD
};

static ERROR UNINSTALL_Activate(struct Uninstall *, APTR);
static ERROR UNINSTALL_ClosingTag(struct Uninstall *, APTR);
static ERROR UNINSTALL_Free(struct Uninstall *, APTR);
static ERROR UNINSTALL_Init(struct Uninstall *, APTR);

static struct ActionArray UninstallActions[] = {
   { AC_Activate,   UNINSTALL_Activate },
   { AC_ClosingTag, UNINSTALL_ClosingTag },
   { AC_Free,       UNINSTALL_Free },
   { AC_Init,       UNINSTALL_Init },
   { NULL, NULL }
};

/*****************************************************************************
** Function prototypes.
*/

static LONG RecursiveDelete(struct Uninstall *, STRING, LONG, WORD);
static void Print(struct Uninstall *, STRING);
static ERROR TagDelete(struct Uninstall *, struct XMLTag *);
static ERROR TagMerge(struct Uninstall *, struct XMLTag *);
static ERROR TagMove(struct Uninstall *, struct XMLTag *);
static ERROR TagRename(struct Uninstall *, struct XMLTag *);
static ERROR TagRun(struct Uninstall *, struct XMLTag *);
static ERROR TagUnregister(struct Uninstall *, struct XMLTag *);

/****************************************************************************/

struct UninstallCommand {
   STRING Tag;
   ERROR (*Routine)(struct Uninstall *, struct XMLTag *);
};

static struct UninstallCommand glTags[] = {
   { "delete",     TagDelete },
   { "merge",      TagMerge },
   { "move",       TagMove },
   { "rename",     TagRename },
   { "run",        TagRun },
   { "unregister", TagUnregister },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &FileSystemMod, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(FileSystemMod, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&UninstallClass, NULL,
      FID_BaseClassID|TLONG, ID_UNINSTALL,
      FID_Version|TFLOAT,    VER_UNINSTALL,
      FID_Name|TSTRING,      "Uninstall",
      FID_Category|TLONG,    CCF_TOOL,
      FID_Actions|TPTR,      UninstallActions,
      FID_Fields|TPTR,       UninstallFields,
      FID_Size|TLONG,        sizeof(struct Uninstall),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (StringsMod)     { Action(AC_Free, StringsMod, NULL);     StringsMod = NULL; }
   if (FileSystemMod)  { Action(AC_Free, FileSystemMod, NULL);  FileSystemMod = NULL; }
   if (UninstallClass) { Action(AC_Free, UninstallClass, NULL); UninstallClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Starts the package removal process.
###END###

*****************************************************************************/

static ERROR UNINSTALL_Activate(struct Uninstall *Self, APTR Void)
{
   struct mtGetXMLString getxml;
   struct XMLTag **xmllist;
   struct DeleteDir *nextdir, *dir;
   OBJECTID desktopid;
   OBJECTPTR file, script;
   LONG count, i;

   if (!Self->prvLocation) return(ObjectError(Self, ERH_Activate, ERR_MissingLocation));

   DPrintF("~Activate()","[Uninstall:%d] %s", Self->Head.UniqueID, Self->prvLocation);

   if (Self->prvXML) {
      Action(AC_Free, Self->prvXML, NULL);
      Self->prvXML = NULL;
   }

   if (CreateObject(ID_XML, NF_CHILD, &Self->prvXML, NULL,
      FID_Location|TSTRING, Self->prvLocation,
      TAGEND) IS ERR_Okay) {

      if (GetFields(Self->prvXML, FID_Tags|TPTR, &xmllist, TAGEND) != ERR_Okay) {
         Action(AC_Free, Self->prvXML, NULL);
         Self->prvXML = NULL;
         StepBack();
         return(ObjectError(Self, ERH_Activate, ERR_GetField));
      }

      /*** Find the <uninstall> tag ***/

      for (count=0; xmllist[count]; count++) {
         if (StrCompare("uninstall", xmllist[count]->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) break;
      }

      if (!xmllist[count]) {
         Print(Self, "No <uninstall> tag was found in the XML file.");
         Action(AC_Free, Self->prvXML, NULL);
         Self->prvXML = NULL;
         StepBack();
         return(ERR_BadData);
      }

      /*** Execute the uninstall tags ***/

      Print(Self, "Removing the package...\n");

      Self->prvTag = xmllist[count]->Child;
      while (Self->prvTag) {
         for (i=0; glTags[i].Tag; i++) {
            if (StrCompare(glTags[i].Tag, Self->prvTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               glTags[i].Routine(Self, Self->prvTag);
               break;
            }
         }

         if (!glTags[i].Tag) {
            DPrintF("~Activate:","[Uninstall] Processing custom tag <%s>", Self->prvTag->Attributes[0].Name);

            /* If we do not recognise the tag, process it as a
            ** scriptable object.
            */

            getxml.Index = Self->prvTag->Index;
            getxml.Flags = NULL;
            if (Action(MT_GetXMLString, Self->prvXML, &getxml) IS ERR_Okay) {
               if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
                  FID_String|TSTRING, getxml.Result,
                  TAGEND) IS ERR_Okay) {
                  Action(AC_Activate, script, NULL);
                  Action(AC_Free, script, NULL);
               }
               else ObjectError(Self, ERH_Activate, ERR_CreateObject);

               FreeMemory(getxml.Result);
            }
            else DPrintF("@Uninstall:","Failed to generate DML string.");

            StepBack();
         }
         Self->prvTag = Self->prvTag->Next;
      }

      /* Delete any directories that have been deferred
      ** to the end of the uninstall process.
      */

      if ((dir = Self->prvDirs)) {
         UBYTE location[400];

         for (i=0; (dir->Path[i]) AND (i < sizeof(location)-1); i++) location[i] = dir->Path[i];
         location[i] = 0;

         for (dir=Self->prvDirs; dir; dir=dir->Next) {
            RecursiveDelete(Self, location, sizeof(location), TRUE);
         }

         /*** Remove directory entries ***/

         for (dir=Self->prvDirs; dir; dir=nextdir) {
            nextdir = dir->Next;
            FreeMemory(dir);
         }
         Self->prvDirs = NULL;
      }

      /*** Delete the uninstall file ***/

      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, Self->prvLocation,
         FID_Flags|TLONG,      FL_APPROXIMATE,
         TAGEND) IS ERR_Okay) {
         Action(MT_DeleteFile, file, NULL);
         Action(AC_Free, file, NULL);
      }

      Action(AC_Free, Self->prvXML, NULL);
      Self->prvXML = NULL;

      if (FastFindObject("desktop", NULL, &desktopid, 1, NULL) IS ERR_Okay) {
         struct acCustom custom;
         custom.Number = 0;
         custom.String = "RefreshIcons";
         ActionMsg(AC_Custom, desktopid, &custom);
      }
   }
   else {
      Print(Self, "The package does not exist.");
      StepBack();
      return(ERR_CreateObject);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Uninstall: ClosingTag
*/

static ERROR UNINSTALL_ClosingTag(struct Uninstall *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      Action(AC_Activate, Self, NULL);
      Action(AC_Free, Self, NULL);
      return(ERR_Okay);
   }
   else return(ERR_Okay);
}

/*****************************************************************************
** Uninstall: Free
*/

static ERROR UNINSTALL_Free(struct Uninstall *Self, APTR Void)
{
   struct DeleteDir *nextdir, *dir;

   if (Self->prvLocation) {
      FreeMemory(Self->prvLocation);
      Self->prvLocation = NULL;
   }

   if (Self->prvXML) {
      Action(AC_Free, Self->prvXML, NULL);
      Self->prvXML = 0;
   }

   for (dir=Self->prvDirs; dir; dir=nextdir) {
      nextdir = dir->Next;
      FreeMemory(dir);
   }
   Self->prvDirs = NULL;

   return(ERR_Okay);
}

/*****************************************************************************
** Uninstall: Init
*/

static ERROR UNINSTALL_Init(struct Uninstall *Self, APTR Void)
{
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:    Location
Synonym: Src
Short:   The location of the uninstall file is specified here.
Status:  Get/Set
Type:    STRING

###DESCRIPTION###
<p>Before intiailising an install object you will need to set the location
of the package that is to be installed.  The package must have a '.install'
extension and be compressed in a format supported by the
<a href="compression.html">Compression</a> class (pkzip is currently
recommended).  If the file is not valid then the install object will fail
when the Activate action is called.</p>

###END###

*****************************************************************************/

static ERROR GET_Location(struct Uninstall *Self, STRING *Value)
{
   *Value = Self->prvLocation;
   return(ERR_Okay);
}

static ERROR SET_Location(struct Uninstall *Self, STRING Value)
{
   LONG i;

   if (Self->prvLocation) { FreeMemory(Self->prvLocation); Self->prvLocation = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->prvLocation, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->prvLocation[i] = Value[i];
         Self->prvLocation[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Output
Short:  Result messages will be sent to the object referred to in this field.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>If this field is set to a valid ObjectID, text messages will be sent to
the object when the print process executes.  This can be helpful for
notifying the user that an error has occurred during the print process.</p>

<p>The object receiving the message must be capable of understanding text
sent via data channels.  In most cases it is recommended that a Text
object is used for this purpose.</p>

###END###

****************************************************************************/

/****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make the object static.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>By default, an Uninstall object will execute itself and then self-destruct
when a closing tag is received.  If you would rather that the object stays
in the system, set this field to TRUE.  If you do this, the only way to get
the Uninstall object to perform is to call the Activate action.</p>

###END###

****************************************************************************/

/*****************************************************************************
** Internal: RecursiveDelete()
** Short:    Recursively tests a location to make sure that it contains no files.
*/

static LONG RecursiveDelete(struct Uninstall *Self, STRING Location, LONG LocationSize, WORD CheckEmpty)
{
   UBYTE buffer[160];
   struct FileInfo *list, *scan;
   LONG j, len, type;
   ERROR error;
   WORD delete;

   for (len=0; Location[len]; len++);

   delete = TRUE;

   if (CheckEmpty IS TRUE) {
      if ((AnalyseLocation(Location, &type) IS ERR_Okay) AND (type IS LOC_DIRECTORY)) {
         if (ReadDirectory(Location, RDF_FILE|RDF_DIRECTORY|RDF_QUALIFY, &list) IS ERR_Okay) {

            for (scan=list; scan; scan=scan->Next) {
               if (scan->Flags & RDF_FILE) {
                  StrFormat(buffer, sizeof(buffer), "Not deleting \"%s\" because it contains files.", Location);
                  Print(Self, buffer);
                  FreeMemory(list);
                  delete = FALSE;
                  break;
               }
            }

            /*** If there are sub-directories, we can delete them if they are empty ***/

            if (delete IS TRUE) {
               for (scan=list; scan; scan=scan->Next) {
                  if (scan->Flags & RDF_DIRECTORY) {
                     for (j=0; scan->Name[j]; j++) Location[len+j] = scan->Name[j];
                     Location[len+j] = 0;
                     if (RecursiveDelete(Self, Location, LocationSize, TRUE) IS FALSE) delete = FALSE;
                     Location[len] = 0;
                  }
               }
            }

            FreeMemory(list);
         }
      }
   }

   if (delete IS TRUE) {
      /*** Tell the user what we are doing ***/

      StrFormat(buffer, sizeof(buffer), "Deleting \"%s\"", Location);
      Print(Self, buffer);

      /*** Delete the file or directory ***/

      if ((error = DeleteFile(Location)) != ERR_Okay) {
         /*** Warn the user if the directory exists but could not be deleted ***/

         if (error != ERR_FileNotFound) {
            StrFormat(buffer, sizeof(buffer), "  Failed to delete \"%s\"", Location);
            Print(Self, buffer);
         }
      }
   }

   return(delete);
}

/*****************************************************************************
** Internal: CommandDelete()
*/

static ERROR TagDelete(struct Uninstall *Self, struct XMLTag *XMLTag)
{
   struct DeleteDir *dir;
   UBYTE location[400];
   LONG i;
   STRING src;
   WORD empty;

   src = NULL;
   empty = FALSE;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("empty", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         empty = TRUE;
      }
   }

   if (!src) return(ERR_Okay);

   if (empty) {
      /*** If the directory needs to be empty, defer the deletion to the end of the uninstall process ***/

      if (!Self->prvDirs) {
         AllocMemory(sizeof(STRING) + StrLength(src) + 1, MEM_STRING, &Self->prvDirs, NULL);
         dir = Self->prvDirs;
      }
      else {
         for (dir=Self->prvDirs; dir->Next; dir = dir->Next);
         AllocMemory(sizeof(STRING) + StrLength(src) + 1, MEM_STRING, &dir->Next, NULL);
         dir = dir->Next;
      }
      for (i=0; src[i]; i++) dir->Path[i] = src[i];
      dir->Path[i] = 0;
      return(ERR_Okay);
   }

   for (i=0; src[i]; i++) location[i] = src[i];
   location[i] = 0;
   RecursiveDelete(Self, location, sizeof(location), empty);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: TagMerge()
*/

static ERROR TagMerge(struct Uninstall *Self, struct XMLTag *XMLTag)
{
   UBYTE buffer[300];
   OBJECTPTR savefile, config;
   struct XMLTag *tag, *childtag;
   STRING file, content, section, item;
   LONG i;
   ERROR error;

   if (!XMLTag->Attributes) return(ERR_Okay);

   DPrintF("~Activate:","Processing <merge> tag.");

   file    = NULL;
   section = NULL;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("file", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         file = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("section", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         section = XMLTag->Attributes[i].Value;
      }
   }

   error = ERR_Okay;
   if ((file) AND (section)) {
      /*** Tell the user what we are doing ***/

      StrFormat(buffer, sizeof(buffer), "Modifying section \"%s\" in file \"%s\"", section, file);
      Print(Self, buffer);

      /*** Open the configuration file ***/

      if ((error = CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
            FID_Location|TSTRING, file,
            TAGEND)) IS ERR_Okay) {

         for (tag=XMLTag->Child; tag; tag=tag->Next) {
            if (StrCompare("item", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               /*** Process an <item> tag ***/

               item = NULL;
               content = NULL;
               for (i=1; i < tag->TotalAttributes; i++) {
                  if (StrCompare("name", tag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     item = tag->Attributes[i].Value;
                  }
                  else DPrintF("@Activate:","[Uninstall] Unsupported <item> attribute: %s.", tag->Attributes[i].Name);
               }

               if (!item) continue;

               /*** Check the tag for content and merge it with the config file ***/

               for (childtag=tag->Child; childtag; childtag=childtag->Next) {
                  if (!childtag->Attributes[0].Name) {
                     ActionTags(MT_AddConfigEntry, config, section, item, childtag->Attributes[0].Value);
                     break;
                  }
               }

               /*** If no content was specified then we delete the item ***/

               if (!childtag) DeleteConfigEntry(config, section, item);
            }
            else if (StrCompare("section", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               for (i=1; i < tag->TotalAttributes; i++) {
                  if (StrCompare("name", tag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     section = tag->Attributes[i].Value;
                     break;
                  }
                  else DPrintF("@Activate:","[Uninstall] Unsupported <section> attribute: %s.", tag->Attributes[i].Name);
               }
            }
            else DPrintF("@Activate:","[Uninstall] Unsupported child tag in <merge> section: <%s>.", tag->Attributes[i].Name);
         }

         /*** Save configuration file changes ***/

         if ((error = CreateObject(ID_FILE, NF_CHILD, &savefile, NULL,
            FID_Location|TSTRING, file,
            FID_Flags|TLONG,      FL_WRITE|FL_NEW,
            TAGEND)) IS ERR_Okay) {
            ActionTags(AC_SaveToObject, config, savefile->UniqueID);
            Action(AC_Free, savefile, NULL);
         }
         else Print(Self, "Failed to save the configuration file changes.");

         Action(AC_Free, config, NULL);
      }
      else Print(Self, "Failed to open the configuration file.");
   }
   else Print(Self, "A <merge> tag in the uninstall file is missing file and/or section attributes.");

   StepBack();
   return(error);
}

/*****************************************************************************
** Internal: TagMove()
*/

static ERROR TagMove(struct Uninstall *Self, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   OBJECTPTR file;
   LONG i;
   STRING src, dest;

   src = NULL;
   dest = NULL;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("dest", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         dest = XMLTag->Attributes[i].Value;
      }
   }

   if ((src) AND (dest)) {
      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, src,
         TAGEND) IS ERR_Okay) {

         /*** Tell the user what we are doing ***/

         StrFormat(buffer, sizeof(buffer), "Moving \"%s\" to \"%s\"", src, dest);
         Print(Self, buffer);

         /*** Move the file ***/

         ActionTags(MT_MoveFile, file, dest);

         Action(AC_Free, file, NULL);
      }
   }
   else DPrintF("@Uninstall:","<move> tag is missing src and/or dest attributes.");

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: TagRename()
*/

static ERROR TagRename(struct Uninstall *Self, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   OBJECTPTR file;
   LONG i;
   STRING src, newname;

   src = NULL;
   newname = NULL;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("newname", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         newname = XMLTag->Attributes[i].Value;
      }
   }

   if ((src) AND (newname)) {
      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, src,
         TAGEND) IS ERR_Okay) {

         /*** Tell the user what we are doing ***/

         StrFormat(buffer, sizeof(buffer), "Renaming \"%s\" to \"%s\"", src, newname);
         Print(Self, buffer);

         /*** Rename the file ***/

         ActionTags(AC_Rename, file, newname);

         Action(AC_Free, file, NULL);
      }
   }
   else DPrintF("@Uninstall:","<rename> tag is missing src and/or newname attributes.");

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: TagRun()
*/

static ERROR TagRun(struct Uninstall *Self, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   OBJECTPTR run;
   STRING src;
   ERROR error;
   LONG i;

   src = NULL;
   if (NewObject(ID_RUN, NF_CHILD, &run, NULL) IS ERR_Okay) {

      for (i=1; i < XMLTag->TotalAttributes; i++) {
         if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetFieldVariable(run, "src", XMLTag->Attributes[i].Value);
            src = XMLTag->Attributes[i].Value;
         }
         else if (StrCompare("args",  XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetField(run, FID_Args, FT_STRING, XMLTag->Attributes[i].Value);
         }
         else if (StrCompare("mode", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetFieldVariable(run, "mode", XMLTag->Attributes[i].Value);
         }
      }

      if (src) {
         StrFormat(buffer, sizeof(buffer), "Running \"%s\"", src);
         Print(Self, buffer);

         if ((error = Action(AC_Init, run, NULL)) IS ERR_Okay) {
            if ((error = Action(AC_Activate, run, NULL)) IS ERR_Okay) {
               Action(AC_Free, run, NULL);
               return(ERR_Okay);
            }
         }
      }
      else error = ERR_Okay;

      Action(AC_Free, run, NULL);
   }
   else error = ERR_NewObject;

   return(error);
}

/*****************************************************************************
** Internal: TagUnregister()
*/

static ERROR TagUnregister(struct Uninstall *Self, struct XMLTag *XMLTag)
{
   UBYTE buffer[300];
   OBJECTPTR file, config;
   STRING classname, location, program;
   LONG i;

   if (!XMLTag->Attributes) return(ERR_Okay);

   DPrintF("~Activate:","Processing <unregister> tag.");

   classname = NULL;
   program   = NULL;
   location  = NULL;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("class", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         classname = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("location", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         location = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("program", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         program = XMLTag->Attributes[i].Value;
      }
   }

   if ((classname) AND (location)) {
      /*** Tell the user what we are doing ***/

      StrFormat(buffer, sizeof(buffer), "Unregistering class %s.", classname);
      Print(Self, buffer);

      /*** Remove the class' entry from the system classes configuration file ***/

      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, "system:config/kernel/classes.cfg",
         TAGEND) IS ERR_Okay) {
         ActionTags(MT_DeleteConfigSection, config, classname);

         if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
            FID_Location|TSTRING, "system:config/kernel/classes.cfg",
            FID_Flags|TLONG,      FL_WRITE|FL_NEW,
            TAGEND) IS ERR_Okay) {
            ActionTags(AC_SaveToObject, config, file->UniqueID);
            Action(AC_Free, file, NULL);
         }

         Action(AC_Free, config, NULL);
      }
      else Print(Self, "Failed to open file system:config/kernel/classes.cfg.");

      /*** Delete the class binary file ***/

      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, location,
         FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
         TAGEND) IS ERR_Okay) {
         Action(MT_DeleteFile, file, NULL);
         Action(AC_Free, file, NULL);
      }
      else Print(Self, "Failed to delete the class binary file.");
   }
   else if (program) {
      /*** Tell the user what we are doing ***/

      StrFormat(buffer, sizeof(buffer), "Unregistering program %s.", program);
      Print(Self, buffer);

      /*** Remove the program entry from the user's program file ***/

      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, "user:config/programs.cfg",
         TAGEND) IS ERR_Okay) {
         ActionTags(MT_DeleteConfigSection, config, program);

         if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
            FID_Location|TSTRING, "user:config/programs.cfg",
            FID_Flags|TLONG,      FL_WRITE|FL_NEW,
            TAGEND) IS ERR_Okay) {
            ActionTags(AC_SaveToObject, config, file->UniqueID);
            Action(AC_Free, file, NULL);
         }

         Action(AC_Free, config, NULL);
      }

      /*** Remove the program entry from the global program file ***/

      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, "system:config/software/programs.cfg",
         TAGEND) IS ERR_Okay) {
         ActionTags(MT_DeleteConfigSection, config, program);

         if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
            FID_Location|TSTRING, "system:config/software/programs.cfg",
            FID_Flags|TLONG,      FL_WRITE|FL_NEW,
            TAGEND) IS ERR_Okay) {
            ActionTags(AC_SaveToObject, config, file->UniqueID);
            Action(AC_Free, file, NULL);
         }

         Action(AC_Free, config, NULL);
      }
      else Print(Self, "Failed to open file system:config/software/programs.cfg.");
   }
   else Print(Self, "An <unregister> tag in the uninstall file is missing class and/or location attributes.");

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: Print()
*/

static void Print(struct Uninstall *Self, STRING Buffer)
{
   struct acDataChannel channel;

   if (Self->OutputID) {
      channel.ObjectID = Self->Head.UniqueID;
      channel.DataType = DATA_TEXT;
      channel.Version  = 1;
      channel.Buffer   = Buffer;
      for (channel.Size=0; Buffer[channel.Size]; channel.Size++);
      channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &channel);
   }
   else DPrintF("Uninstall:", Buffer);
}

/*****************************************************************************
** Internal: DeleteConfigEntry()
*/

static void DeleteConfigEntry(OBJECTPTR Config, STRING Section, STRING Item)
{
   struct ConfigEntry *entries;
   LONG i, amtentries;

   if (GetFields(Config, FID_Entries|TPTR,     &entries,
                         FID_AmtEntries|TLONG, &amtentries,
                         TAGEND) IS ERR_Okay) {
      for (i=0; i < amtentries; i++) {
         if (StrCompare(Section, entries[i].Section, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            if (StrCompare(Item, entries[i].Item, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               if (ActionTags(MT_DeleteConfigEntry, Config, i) IS ERR_Okay) return;
               else break;
            }
         }
      }
   }

   DPrintF("@Uninstall:","Failed to delete config item \"%s\" in section \"%s\"", Item, Section);
}
